{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "[Day 7](https://adventofcode.com/2019/day/7) puzzle input is an IntCode program that that produces an output based on its input values (mine is [here](puzzle_input/day7.txt)). Part 1 involves figuring out which inputs produce the largest output value. Part 2 is similar, except you configure the IntCode cpu slightly differently and run it many times - to find the largest output. AoC describes it better than I can :-)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "880726\n"
     ]
    }
   ],
   "source": [
    "opcode_add = 1\n",
    "opcode_mul = 2\n",
    "opcode_read = 3\n",
    "opcode_write = 4\n",
    "opcode_jump_true = 5\n",
    "opcode_jump_false = 6\n",
    "opcode_lt = 7\n",
    "opcode_eq = 8\n",
    "opcode_rebase = 9\n",
    "\n",
    "opcode_terminate = 99\n",
    "\n",
    "mode_position = 0\n",
    "mode_immediate = 1\n",
    "mode_relative = 2\n",
    "\n",
    "class IntCodeCpu:\n",
    "    def __init__(self, memory_image):\n",
    "        self.memory = [ x for x in memory_image ] # copy memory image, in case it's reused\n",
    "        self.stalled = True\n",
    "        self.input_buffer = None\n",
    "        self.output_buffer = None\n",
    "        self.pc = 0\n",
    "        self.initialise_opcodes()\n",
    "        self.offset = 0\n",
    "        self.halted = True\n",
    "        \n",
    "    def start(self, input_buffer, output_buffer, noun=None, verb=None):\n",
    "        self.input_buffer = input_buffer\n",
    "        self.output_buffer = output_buffer\n",
    "\n",
    "        if noun:\n",
    "            self.memory[1] = noun\n",
    "        if verb:\n",
    "            self.memory[2] = verb\n",
    "\n",
    "        self.stalled = False\n",
    "        self.halted = False\n",
    "            \n",
    "        return self.run()\n",
    "            \n",
    "            \n",
    "    def run(self):\n",
    "        instr = self.memory[self.pc]\n",
    "\n",
    "        while int(instr) != opcode_terminate and not self.stalled:\n",
    "            (op, modes) = self.decode_instr(instr)\n",
    "            self.pc = op(modes)\n",
    "            instr = self.memory[self.pc]\n",
    "        \n",
    "        if not self.stalled:\n",
    "            self.halted = True\n",
    "            \n",
    "        return self.memory[0]\n",
    "\n",
    "    #-HELPERS-----------------------------\n",
    "    def try_pop_mode(self, modes):\n",
    "        if len(modes) == 0:\n",
    "            return 0\n",
    "    \n",
    "        return modes.pop()\n",
    "    \n",
    "    def resize_memory(self, target_addr):\n",
    "        self.memory += ([0] * (1 + target_addr - len(self.memory)))\n",
    "    \n",
    "    #-DECODE-INSTRUCTIONS-----------------\n",
    "    def initialise_opcodes(self):\n",
    "        self.opcodes = {\n",
    "            opcode_add: self.op_add,\n",
    "            opcode_mul: self.op_mul,\n",
    "            opcode_read: self.op_read,\n",
    "            opcode_write: self.op_write,\n",
    "            opcode_jump_true: self.op_jump_true,\n",
    "            opcode_jump_false: self.op_jump_false,\n",
    "            opcode_lt: self.op_lt,\n",
    "            opcode_eq: self.op_eq,\n",
    "            opcode_rebase: self.op_rebase\n",
    "        }    \n",
    "    \n",
    "    def decode_instr(self, instr):\n",
    "        instr = str(instr)\n",
    "        opcode = self.decode_op(instr)\n",
    "        modes = self.decode_modes(instr)\n",
    "\n",
    "        if not (opcode in self.opcodes):\n",
    "            raise Exception(f\"Invalid opcode {opcode}\")\n",
    "\n",
    "        return (self.opcodes[opcode], modes)    \n",
    "    \n",
    "    def decode_op(self, instr):\n",
    "        if len(instr) > 2:\n",
    "            return int(instr[-2:])\n",
    "        return int(instr)\n",
    "    \n",
    "    def decode_modes(self, instr):\n",
    "        if len(instr) > 2:\n",
    "            return [ int(d) for d in instr[:-2]]\n",
    "        return []\n",
    "    \n",
    "    #-MICRO-OPS---------------------------\n",
    "    def uop_read(self, value, mode):\n",
    "        if mode == mode_position:\n",
    "            if value >= len(self.memory):\n",
    "                self.resize_memory(value)        \n",
    "\n",
    "            return int(self.memory[value])\n",
    "\n",
    "        elif mode == mode_relative:\n",
    "            if self.offset + value >= len(self.memory):\n",
    "                self.resize_memory(self.offset + value)     \n",
    "\n",
    "            return int(self.memory[self.offset + value])\n",
    "\n",
    "        elif mode == mode_immediate:\n",
    "            return int(value)\n",
    "\n",
    "        else:\n",
    "            raise Exception(\"UNKNOWN MODE\")\n",
    "\n",
    "    def uop_write(self, dst, value, mode):\n",
    "        if mode == mode_position:\n",
    "            if dst >= len(self.memory):\n",
    "                self.resize_memory(dst)\n",
    "            self.memory[dst] = value\n",
    "\n",
    "        elif mode == mode_relative:\n",
    "            if self.offset + dst >= len(self.memory):\n",
    "                self.resize_memory(self.offset + dst)     \n",
    "\n",
    "            self.memory[self.offset + dst] = value\n",
    "\n",
    "        elif mode == mode_immediate:\n",
    "            raise Exception(f\"cannot write {value} to literal {dst}\")\n",
    "\n",
    "\n",
    "    def uop_cond_jump(self, modes, cond):\n",
    "        param_mode = self.try_pop_mode(modes)\n",
    "        param_raw = int(self.memory[self.pc + 1])\n",
    "        param = self.uop_read(param_raw, param_mode)\n",
    "\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest_raw = int(self.memory[self.pc + 2])\n",
    "        dest = self.uop_read(dest_raw, dest_mode)\n",
    "\n",
    "        if cond(param):\n",
    "            return dest\n",
    "\n",
    "        return self.pc + 3\n",
    "\n",
    "    def uop_cmp(self, modes, cmp):\n",
    "\n",
    "        param0_mode = self.try_pop_mode(modes)\n",
    "        param0_raw = int(self.memory[self.pc + 1])\n",
    "        param0 = self.uop_read(param0_raw, param0_mode)\n",
    "\n",
    "        param1_mode = self.try_pop_mode(modes)\n",
    "        param1_raw = int(self.memory[self.pc + 2])\n",
    "        param1 = self.uop_read(param1_raw, param1_mode)\n",
    "\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        if cmp(param0, param1):\n",
    "            self.uop_write(dest, 1, dest_mode)\n",
    "        else:\n",
    "            self.uop_write(dest, 0, dest_mode)\n",
    "\n",
    "        return self.pc + 4\n",
    "        \n",
    "    \n",
    "    #-OPCODES-----------------------------\n",
    "    def op_add(self, modes):\n",
    "        arg0_mode = self.try_pop_mode(modes)\n",
    "        arg1_mode = self.try_pop_mode(modes)\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "\n",
    "        arg0_raw = int(self.memory[self.pc + 1])\n",
    "        arg1_raw = int(self.memory[self.pc + 2])\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        arg0 = self.uop_read(arg0_raw, arg0_mode)\n",
    "        arg1 = self.uop_read(arg1_raw, arg1_mode)\n",
    "\n",
    "        self.uop_write(dest, str(int(arg0) + int(arg1)), dest_mode)\n",
    "        return self.pc + 4\n",
    "\n",
    "    def op_mul(self, modes):\n",
    "        arg0_mode = self.try_pop_mode(modes)\n",
    "        arg1_mode = self.try_pop_mode(modes)\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "\n",
    "        arg0_raw = int(self.memory[self.pc + 1])\n",
    "        arg1_raw = int(self.memory[self.pc + 2])\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        arg0 = self.uop_read(arg0_raw, arg0_mode)\n",
    "        arg1 = self.uop_read(arg1_raw, arg1_mode)\n",
    "\n",
    "        self.uop_write(dest, str(int(arg0) * int(arg1)), dest_mode)    \n",
    "        return self.pc + 4\n",
    "\n",
    "\n",
    "    def op_read(self, modes):\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest = int(self.memory[self.pc + 1])\n",
    "        \n",
    "        # if the input buffer is empty, we should \"stall\" and \n",
    "        # resume later\n",
    "        if not self.input_buffer:\n",
    "            self.stalled = True\n",
    "            return self.pc\n",
    "        \n",
    "        val = self.input_buffer.pop()\n",
    "\n",
    "        self.uop_write(dest, str(val), dest_mode)\n",
    "\n",
    "        return self.pc + 2\n",
    "\n",
    "    def op_write(self, modes):\n",
    "        src_mode = self.try_pop_mode(modes)\n",
    "        src_raw = int(self.memory[self.pc + 1])\n",
    "        src = self.uop_read(src_raw, src_mode)\n",
    "\n",
    "        self.output_buffer.append(src)\n",
    "\n",
    "        return self.pc + 2\n",
    "\n",
    "    def op_jump_true(self, modes):\n",
    "        return self.uop_cond_jump(modes, lambda x: x != 0)\n",
    "\n",
    "\n",
    "    def op_jump_false(self, modes):\n",
    "        return self.uop_cond_jump(modes, lambda x: x == 0)\n",
    "\n",
    "    def op_lt(self, modes):\n",
    "        return self.uop_cmp(modes, lambda x, y: x < y)\n",
    "\n",
    "    def op_eq(self, modes):\n",
    "        return self.uop_cmp(modes, lambda x, y: x == y)\n",
    "\n",
    "    def op_rebase(self, modes):    \n",
    "        param_mode = self.try_pop_mode(modes)\n",
    "        param_raw = int(self.memory[self.pc + 1])\n",
    "        param = self.uop_read(param_raw, param_mode)\n",
    "\n",
    "        self.offset += param\n",
    "\n",
    "        return self.pc + 2 \n",
    "        \n",
    "\n",
    "def calculate_signal(phase_settings_tuple, program):\n",
    "    phase_settings = list(phase_settings_tuple)\n",
    "    input_signal = 0\n",
    "    output_buffer = []\n",
    "    \n",
    "    cpus = [ IntCodeCpu(program) for phase in phase_settings ]\n",
    "    \n",
    "    all_stalled = True\n",
    "    signal = 0\n",
    "\n",
    "    for cpu in cpus:\n",
    "        input_buffer = [ signal ]\n",
    "        if phase_settings:\n",
    "            input_buffer.append(phase_settings.pop(0))\n",
    "        cpu.start(input_buffer, [])\n",
    "        signal = cpu.output_buffer.pop(0)\n",
    "        \n",
    "    return signal\n",
    "    \n",
    "from itertools import permutations\n",
    "\n",
    "def max_signal(raw_code):\n",
    "    code = raw_code.split(\",\")\n",
    "    phase_setting_combinations = permutations([0,1,2,3,4])\n",
    "    all_signals = [ calculate_signal(phase_settings, code) for phase_settings in phase_setting_combinations ]\n",
    "    return max(all_signals)\n",
    "\n",
    "test_input1 = \"3,15,3,16,1002,16,10,16,1,16,15,15,4,15,99,0,0\"\n",
    "test_input2 = \"3,23,3,24,1002,24,10,24,1002,23,-1,23,101,5,23,23,1,24,23,23,4,23,99,0,0\"\n",
    "test_input3 = \"3,31,3,32,1002,32,10,32,1001,31,-2,31,1007,31,0,33,1002,33,7,33,1,33,31,31,1,32,31,31,4,31,99,0,0,0\"\n",
    "\n",
    "assert 43210 == max_signal(test_input1)\n",
    "assert 54321 == max_signal(test_input2)\n",
    "assert 65210 == max_signal(test_input3)\n",
    "\n",
    "puzzle_input = open(\"puzzle_input/day7.txt\").read()\n",
    "\n",
    "print(max_signal(puzzle_input))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "4931744\n"
     ]
    }
   ],
   "source": [
    "def calculate_looped_signal(phase_settings_tuple, program):\n",
    "    phase_settings = list(phase_settings_tuple)\n",
    "    input_signal = 0\n",
    "    output_buffer = []\n",
    "    \n",
    "    cpus = [ IntCodeCpu(program) for phase in phase_settings ]\n",
    "    \n",
    "    all_stalled = True\n",
    "    signal = 0\n",
    "\n",
    "    while all_stalled:\n",
    "        for cpu in cpus:\n",
    "            input_buffer = [ signal ]\n",
    "            if phase_settings:\n",
    "                input_buffer.append(phase_settings.pop(0))\n",
    "            cpu.start(input_buffer, [])\n",
    "            signal = cpu.output_buffer.pop(0)\n",
    "            all_stalled = all_stalled and cpu.stalled\n",
    "    \n",
    "    return signal\n",
    "    \n",
    "    \n",
    "def max_looped_signal(raw_code):\n",
    "    code = raw_code.split(\",\")\n",
    "    phase_setting_combinations = permutations([5,6,7,8,9])\n",
    "    all_signals = [ calculate_looped_signal(phase_settings, code) for phase_settings in phase_setting_combinations ]\n",
    "    return max(all_signals)\n",
    "\n",
    "\n",
    "assert 139629729 == max_looped_signal(\"3,26,1001,26,-4,26,3,27,1002,27,2,27,1,27,26,27,4,27,1001,28,-1,28,1005,28,6,99,0,0,5\")\n",
    "assert 18216 == max_looped_signal(\"3,52,1001,52,-5,52,3,53,1,52,56,54,1007,54,5,55,1005,55,26,1001,54,-5,54,1105,1,12,1,53,54,53,1008,54,0,55,1001,55,1,55,2,53,55,53,4,53,1001,56,-1,56,1005,56,6,99,0,0,0,0,10\")\n",
    "\n",
    "puzzle_input = open(\"puzzle_input/day7.txt\").read()\n",
    "print(max_looped_signal(puzzle_input))"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
